//$Id: ufl.cpp,v 1.10 2003/12/03 01:35:55 magh Exp $

/* ------------------------------------------------------------------------
 Author: Matthew Galati (magh@lehigh.edu)

 (c) Copyright 2003 Lehigh University. All Rights Reserved.

 This software is licensed under the Common Public License. Please see 
 accompanying file for terms.    
---------------------------------------------------------------------------*/

/*-------------------------------------------------------------------------*/
#include "ufl.hpp"
#include <iostream>
#include <fstream>
using namespace std;

/*-------------------------------------------------------------------------*/
#ifdef COIN_USE_CLP
#include "OsiClpSolverInterface.hpp"
typedef OsiClpSolverInterface OsiXxxSolverInterface;
#include "SbbModel.hpp"
#endif

#ifdef COIN_USE_OSL
#include "OsiOslSolverInterface.hpp"
typedef OsiOslSolverInterface OsiXxxSolverInterface;
#include "ekk_c_api.h"
#endif

#ifdef COIN_USE_CPX
#include "OsiCpxSolverInterface.hpp"
typedef OsiCpxSolverInterface OsiXxxSolverInterface;
#endif

#include "CoinHelperFunctions.hpp"
#include "CoinPackedVector.hpp"

#include "CglGomory.hpp"
#include "CglKnapsackCover.hpp"
#include "CglLiftAndProject.hpp"
#include "CglOddHole.hpp"
#include "CglProbing.hpp"
#include "CglSimpleRounding.hpp"

/*-------------------------------------------------------------------------*/
UFL::UFL(const char * datafile)
  : si(0), trans_cost(0), fixed_cost(0), demand(0), 
    M(0), N(0), total_demand(0.0){
  read_data(datafile);
}

/*-------------------------------------------------------------------------*/
UFL::~UFL(){
  if(trans_cost != 0){
    delete [] trans_cost; trans_cost = 0;
  }
  if(fixed_cost != 0){
    delete [] fixed_cost; fixed_cost = 0;
  }
  if(demand != 0){
    delete [] demand; demand = 0;
  }
  if(integer_vars != 0){
    delete [] integer_vars; integer_vars = 0;
  }
  if(si != 0){ 
    delete si; si = 0; 
  }
}

/*-------------------------------------------------------------------------*/
OsiSolverInterface* UFL::initialize_solver(){
  OsiXxxSolverInterface* si = new OsiXxxSolverInterface();

#ifdef COIN_USE_OSL
  //this shuts off the OSL messages, turn them on to debug
  ekk_messagesPrintOff(si->getModelPtr(), 1, 3048);
  ekk_messagePrintOn(si->getModelPtr(), 106);
#endif

  return si;
}

/*-------------------------------------------------------------------------*/
void UFL::create_model(){
  /*
    First, we must initialize the OSI solver interface. The solver choice
    is determined by the -DCOIN_USE_XXX in the Makefile.
  */
  si = initialize_solver();

  /*
    Next, we must define the columns (variables) in the problem. The 
    functions xindex(i,j) and yindex(j) define the order of the 
    columns. First we list the x indices (i = 0, ..., M-1, j = 0, 
    ..., N-1), then the y indices (j = 0, ..., N-1). We will also store 
    the indices of those columns which are integral (the y variables).
  */
  int n_cols = (M * N) + N;
  double * objective    = new double[n_cols];//the objective coefficients
  double * col_lb       = new double[n_cols];//the column lower bounds
  double * col_ub       = new double[n_cols];//the column upper bounds
  integer_vars          = new int[N];        //the indices of the integral vars

  /*
    Here we use some COIN helper functions which are defined in:
    Coin/include/CoinHelperFunctions.hpp
  */
  CoinIotaN(integer_vars, N, M * N);
  CoinFillN(col_lb, n_cols, 0.0);

  int i, j, index;
  for(i = 0; i < M; i++){
    for(j = 0; j < N; j++){
      index            = xindex(i,j);
      objective[index] = trans_cost[index] / demand[i];
      col_ub[index]    = demand[i];
    }
  }
  CoinFillN(col_ub + (M*N), N, 1.0);
  CoinDisjointCopyN(fixed_cost, N, objective + (M * N));

  /*
    Next we must define the set of rows (constraints). In order to do
    this we will store the information in a CoinPackedMatrix.
 
    There are several ways to construct a CoinPackedMatrix. In this 
    implementation, we will define the matrix as row-ordered.
  */

  int n_rows = M + N;
  double * row_lb = new double[n_rows]; //the row lower bounds
  double * row_ub = new double[n_rows]; //the row upper bounds

  /*
    There are M demand constraints  (= d[i]), LB[i] = UB[i] = d[i]
    There are N linking constraints (<= 0.0), LB[i] = -infinity, UB = 0.0
    
    Since different solvers define infinity differently, we use the 
    OSI access function getInfinity() to fetch the proper value.
  */

  CoinDisjointCopyN(demand, M, row_lb);
  CoinDisjointCopyN(demand, M, row_ub);
  CoinFillN(row_lb + M, N, -1 * si->getInfinity());
  CoinFillN(row_ub + M, N, 0.0);
  
  CoinPackedMatrix * matrix =  new CoinPackedMatrix(false,0,0);
  matrix->setDimensions(0, n_cols);
  
  //demand constraints: 
  //sum{j in N} x[i][j] = d[i], for i in M
  for(i = 0; i < M; i++){
    CoinPackedVector row;
    for(j = 0; j < N; j++)
      row.insert(xindex(i,j), 1.0);
    matrix->appendRow(row);
  }

  //linking constraints: 
  //sum{i in M} x[i][j] - total_demand * y[j] <= 0, for j in N
  for(j = 0; j < N; j++){
    CoinPackedVector row;
    row.insert(yindex(j), -1.0 * total_demand);
    for(i = 0; i < M; i++)
      row.insert(xindex(i,j), 1.0);
    matrix->appendRow(row);
  }  
  
  si->loadProblem(*matrix, col_lb, col_ub, objective, row_lb, row_ub);

  /*
    Now free up the memory. Notice, the difference between using OSI's
    loadProblem vs assignProblem. With assignProblem, OSI assumes 
    ownership of the arguments and deletes the memory for the user.
  */

  if(objective != 0)   { delete [] objective; objective = 0; }
  if(col_lb != 0)      { delete [] col_lb; col_lb = 0; }
  if(col_ub != 0)      { delete [] col_ub; col_ub = 0; }
  if(row_lb != 0)      { delete [] row_lb; row_lb = 0; }
  if(row_ub != 0)      { delete [] row_ub; row_ub = 0; }
  if(matrix != 0)      { delete matrix; matrix = 0; }
}

/*-------------------------------------------------------------------------*/
void UFL::solve_model(ostream * os){
  /* 
     In this method we will solve UFL exactly with branch and bound.
     Some of the solvers have built-in branchAndBound. In this case,
     we can simply call si->branchAndBound(). However, some solvers 
     (e.g., CLP), do not have a built-in branch and bound. So, we will
     use the SBB module. 

     COIN NOTE: Rather than having the base implementation OsiSolverInterface::
     branchAndBound() be pure virtual, should we make the default to implement 
     using SBB? The negative is that OSI would then depend on SBB.
  */

  si->setInteger(integer_vars, N);

#ifdef COIN_USE_CLP
  SbbModel model(*si);
  model.branchAndBound();
  if(model.isProvenOptimal()){
    const double * solution = model.getColSolution();
    const double * objCoeff = model.getObjCoefficients();
    print_solution(solution, objCoeff, os);
  }
  else
    cerr << "B&B failed to find optimal" << endl;
  return;
#endif

#ifdef COIN_USE_OSL
  si->branchAndBound();
  if(si->isProvenOptimal()){
    const double * solution = si->getColSolution();
    const double * objCoeff = si->getObjCoefficients();
    print_solution(solution, objCoeff, os);
  }
  else
    cerr << "B&B failed to find optimal" << endl;
  return;
#endif

  throw CoinError("Sorry, I have only tested CLP and OSL",
		  "solve_model", "UFL");
}

/*-------------------------------------------------------------------------*/
double UFL::simulate_root(ostream * os){

  /* 
     Get a lower bound by solving the initial LP relaxation.
  */

  si->initialSolve();  
  double current_lb = si->getObjValue();
  double current_ub = si->getInfinity();
  double gap, new_lb, new_ub, improvement = 1.0;
  int iterations = 0;
  int newcuts = 1;
  
  while((newcuts || improvement > 0.01) && iterations <= 100){

    //find an upper bound using a simple rounding heuristic
    new_ub = find_upperbound();
    if(new_ub < current_ub) 
      current_ub = new_ub;
    gap = 100 * (current_ub - current_lb) / current_ub;

    //generate new logical cuts
    newcuts = generate_cuts();

    (*os).setf(ios::fixed);
    (*os).precision(3);
    (*os) << "ITERATION " << iterations << "\tLB = " << current_lb
	  << "\tUB = " << current_ub << "\tGAP = " << gap << "%"    
	  << "\tCUTS L: " << newcuts;

    newcuts += generate_cuts_CGL();
    
    //resolve to produce a new lower bound
    si->resolve();
    new_lb = si->getObjValue();
    improvement = (new_lb - current_lb) / new_lb;
    current_lb = new_lb;
    iterations++;
  }
  gap = 100 * (current_ub - current_lb)/current_ub;

  //print the best solution found in the root node
  const double * solution = si->getColSolution();
  const double * objCoeff = si->getObjCoefficients();
  print_solution(solution, objCoeff, os);

  return gap;

}

/*-------------------------------------------------------------------------*/
double UFL::find_upperbound(){
	
  const double * sol   = si->getColSolution();
  const double * coeff = si->getObjCoefficients();
  double objective_value = si->getObjValue();
 
  int index;	
  for(int j = 0; j < N; j++){
    index = yindex(j);
    const double xi = sol[index];
		
    //if y_j is fractional, round it up to 1
    if (xi > tolerance && xi < 1 - tolerance)
      objective_value += coeff[index] * (1.0 - sol[index]);
  }

  return objective_value;
}

/*-------------------------------------------------------------------------*/
int UFL::generate_cuts(){

  /* 
     Look for disaggregated logical constraints:
     x_ij - (d_j * y_j) <= 0, for i in M (client), j in N (depot)
  */

  const double* sol = si->getColSolution();
  int newcuts = 0;
	
  int i, j, xind, yind;
  for(i = 0; i < M; i++){
    for(j = 0; j < N; j++){
      xind = xindex(i,j);
      yind = yindex(j);

      if(sol[xind] - (demand[i] * sol[yind]) > tolerance){

	//Then, we have a violated constraint - so add it.
	CoinPackedVector cut;
	cut.insert(xind,  1.0);
	cut.insert(yind, -1.0 * demand[i]);
	si->addRow(cut,  -1.0 * si->getInfinity(), 0.0);
	newcuts++;
				
      }

    }
  }

  return newcuts;
}

/*-------------------------------------------------------------------------*/
int UFL::generate_cuts_CGL(ostream * os){
  OsiCuts cutlist;

  int n_gomory, n_knap, n_round, n_odd, n_probe, n_liftproj;
  int n_cuts = 0;

  si->setInteger(integer_vars, N);

  //create CGL gomory cuts
  CglGomory * gomory = new CglGomory;
  gomory->setLimit(100);
  gomory->generateCuts(*si, cutlist);
  n_gomory = cutlist.sizeRowCuts() - n_cuts;
  n_cuts = cutlist.sizeRowCuts();

  //create CGL knapsack cuts
  CglKnapsackCover * knapsack = new CglKnapsackCover;
  knapsack->generateCuts(*si, cutlist);
  n_knap = cutlist.sizeRowCuts() - n_cuts;
  n_cuts = cutlist.sizeRowCuts();

  //create CGL simple rounding cuts
  CglSimpleRounding * rounding = new CglSimpleRounding;
  rounding->generateCuts(*si, cutlist);
  n_round = cutlist.sizeRowCuts() - n_cuts;
  n_cuts = cutlist.sizeRowCuts();

  //create CGL odd hole cuts
  CglOddHole * oddhole = new CglOddHole;
  oddhole->generateCuts(*si, cutlist);
  n_odd = cutlist.sizeRowCuts() - n_cuts;
  n_cuts = cutlist.sizeRowCuts();

  //create CGL probing cuts
  CglProbing * probe = new CglProbing;
  probe->generateCuts(*si, cutlist);
  n_probe = cutlist.sizeRowCuts() - n_cuts;
  n_cuts = cutlist.sizeRowCuts();

  (*os) << " G: " << n_gomory
	<< " K: " << n_knap
	<< " R: " << n_round
	<< " O: " << n_odd
	<< " P: " << n_probe << endl;

  //apply the cuts to the interface
  si->applyCuts(cutlist);
  
  //clean up
  delete gomory;
  delete knapsack;
  delete rounding;
  delete oddhole;
  delete probe;

  return cutlist.sizeRowCuts();
}

/*-------------------------------------------------------------------------*/
void UFL::print_solution(const double * solution, 
			 const double * objCoeff, 
			 ostream * os){

  (*os).setf(ios::fixed);
  (*os).precision(5);
  int n_cols = (M * N) + N, c = 0, i, j;
  double objective = 0.0;
  for(i = 0; i < M; i++){
    for(j = 0; j < N; j++){
      if(solution[c] > tolerance){
	objective += objCoeff[c] * solution[c];
	(*os) << "c : " << c << " x[" << i << "," << j << "] = " 
	      << solution[c] << endl;
      }
      c++;
    }
  }
  for(j = 0; j < N; j++){
    if(solution[c] > tolerance){
      objective += objCoeff[c] * solution[c];
      (*os) << "c : " << c << " y[" << j << "] = " 
	    << solution[c] << endl;
    }
    c++;
  }

  (*os) << "Objective = " << objective << endl;
}

/*-------------------------------------------------------------------------*/
void UFL::read_data(const char * datafile) {

  /*
    http://mscmga.ms.ic.ac.uk/jeb/orlib/uncapinfo.html

    The format of these data files is:
    number of potential warehouse locations (n), number of customers (m)
    for each potential warehouse location j (j = 1,...,n): 
    capacity, fixed cost
    for each customer i (i = 1,...,m): 
    demand, 
    for each warehouse j (j = 1,...,n)
    cost of allocating all of the demand of i to warehouse j
  */

  ifstream is(datafile);
  if (!is) {
    cout << "ERROR I/O: Failure to open: " << datafile << endl;
    abort();
  }

  //number of depots,clients 
  is >> N >> M;
  
  trans_cost = new double[M * N];
  fixed_cost = new double[N];
  demand     = new double[M];

  double dummy; //we are not using the capacity information
  int i, j, index;

  for(j = 0; j < N; j++)
    is >> dummy >> fixed_cost[j];

  for(i = 0; i < M; i++){
    is >> demand[i];
    total_demand += demand[i];
    for(j = 0; j < N; j++){
      int index = xindex(i,j);
      is >> trans_cost[index];
    }
  }

  is.close();
}
